import { BaseMergerStream, Readable } from '@tanker/stream-base';
export { Duplex, Readable, ResizerStream, Transform, Uint8Buffer, Writable } from '@tanker/stream-base';
import { InvalidArgument } from '@tanker/errors';
import FilePonyfill from '@tanker/file-ponyfill';
import FileReader from '@tanker/file-reader';

const defaultMime = 'application/octet-stream';
const getConstructorName = constructor => {
  if (constructor === ArrayBuffer) return 'ArrayBuffer';else if (constructor === Uint8Array) return 'Uint8Array';else if (constructor === File || constructor === FilePonyfill) // must be before Blob
    return 'File'; // else if (constructor === Blob)

  return 'Blob';
};
const converters = {
  ArrayBuffer: uint8array => uint8array.buffer.byteLength === uint8array.length ? uint8array.buffer : new Uint8Array(uint8array).buffer,
  Uint8Array: uint8array => uint8array,
  Blob: (uint8array, {
    mime
  }) => new Blob([uint8array], {
    type: mime || defaultMime
  }),
  File: (uint8array, {
    mime,
    name,
    lastModified
  }) => new FilePonyfill([uint8array], name || '', {
    type: mime || defaultMime,
    lastModified: lastModified || Date.now()
  })
};
class MergerStream extends BaseMergerStream {
  // <T: A | B> is interpreted as 'T must be A or B at all times', not 'T cannot exist unless it is either A or B and should just not change after instantiation'
  // $FlowIssue https://github.com/facebook/flow/issues/7449
  constructor({
    type = Uint8Array,
    mime,
    name,
    ...otherOptions
  } = {}) {
    if (![ArrayBuffer, Blob, File, FilePonyfill, Uint8Array].some(klass => type === klass)) throw new InvalidArgument('options.type', 'class in [ArrayBuffer, Blob, File, FilePonyfill, Uint8Array]', type);
    if (mime && typeof mime !== 'string') throw new InvalidArgument('options.mime', 'string', mime);
    if (name && typeof name !== 'string') throw new InvalidArgument('options.name', 'string', name);
    super({ ...otherOptions,
      type,
      mime,
      name,
      converter: converters[getConstructorName(type)]
    });
  }

}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

class SlicerStream extends Readable {
  constructor(options) {
    if (!options || typeof options !== 'object' || options instanceof Array) throw new InvalidArgument('options', 'object', options);
    const {
      source: _source,
      outputSize
    } = options;
    if ([ArrayBuffer, Uint8Array, Blob, File].every(constructor => !(_source instanceof constructor))) throw new InvalidArgument('options.source', 'either an ArrayBuffer, a Uint8Array, a Blob, or a File', _source);
    if (outputSize && typeof outputSize !== 'number') throw new InvalidArgument('options.outputSize', 'number', outputSize);
    super({
      // buffering a single output chunk
      objectMode: true,
      highWaterMark: 1
    });

    _defineProperty(this, "_mode", void 0);

    _defineProperty(this, "_source", void 0);

    _defineProperty(this, "_outputSize", void 0);

    _defineProperty(this, "_readingState", void 0);

    _defineProperty(this, "_fileReader", void 0);

    _defineProperty(this, "_read", () =>
    /* size: number */
    {
      if (this._mode === 'binary') {
        this._readBinary();
      } else {
        this._readFile();
      }
    });

    _defineProperty(this, "_initBinaryMode", source => {
      this._mode = 'binary';
      this._source = source instanceof ArrayBuffer ? new Uint8Array(source) : source;
      this._readingState = {
        byteSize: this._source.length,
        byteIndex: 0
      };
    });

    _defineProperty(this, "_readBinary", () => {
      const {
        byteSize,
        byteIndex: startIndex
      } = this._readingState;
      const endIndex = Math.min(startIndex + this._outputSize, byteSize); // $FlowIKnow we know _source is an Uint8Array

      const bytes = this._source.subarray(startIndex, endIndex);

      const pushMore = this.push(bytes);

      if (endIndex >= byteSize) {
        this.push(null);
        return;
      }

      this._readingState.byteIndex = endIndex;

      if (pushMore) {
        this._readBinary();
      }
    });

    _defineProperty(this, "_initFileMode", source => {
      this._mode = 'file';
      this._source = source;
      this._fileReader = new FileReader(source);
    });

    _defineProperty(this, "_readFile", async () => {
      try {
        const buffer = await this._fileReader.readAsArrayBuffer(this._outputSize);
        const length = buffer.byteLength;

        if (length === 0) {
          this.push(null);
          return;
        }

        this.push(new Uint8Array(buffer));

        if (length < this._outputSize) {
          this.push(null);
        }
      } catch (error) {
        this.emit('error', error);
      }
    });

    this._outputSize = outputSize || 5 * 1024 * 1024; // 5MB

    if (_source instanceof ArrayBuffer || _source instanceof Uint8Array) {
      this._initBinaryMode(_source);
    } else {
      this._initFileMode(_source);
    }
  }

}

export { MergerStream, SlicerStream };
